#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# This is a modification of Campbell Barton's script io_anim_camera.py.
# The output is a camera animation file for Octane Render.

# Octane Render Camera (ORC) animation

# @author: Campbell Barton, Michael Martin

import datetime
import os
import bpy
import bpy.path as bpath
from mathutils import Vector
from bpy.types import Panel, Operator
from bpy.props import StringProperty, IntProperty, BoolProperty, EnumProperty
from octanerender.utils import *

# ---------------------------------------------------------------------------
def writeOrcLine(obj, orcFile, octane_render):

    unitFactor = {0:0.001,1:0.01,2:0.1,3:1,4:10,5:100,6:1000,
                7:0.0254,8:0.3048,9:0.9144,10:201.168,11:1609.344}[int(octane_render.unit_size)]

    matrix = obj.matrix_world.copy()
    if octane_render.export_ROTX90 == True:
        position = Vector(rotate90x(matrix_vect(matrix,[0.0,0.0,0.0,1.0]))) * unitFactor
        target = Vector(rotate90x(matrix_vect(matrix,[0.0,0.0,-1.0,1.0]))) * unitFactor
        up = Vector(rotate90x(matrix_vect(matrix,[0.0,1.0,0.0,0.0]))) * unitFactor
    else:
        position = Vector(matrix_vect(matrix,[0.0,0.0,0.0,1.0])) * unitFactor
        target = Vector(matrix_vect(matrix,[0.0,0.0,-1.0,1.0])) * unitFactor
        up = Vector(matrix_vect(matrix,[0.0,1.0,0.0,0.0])) * unitFactor

    camera_settings = []
    camera_settings.append('%f, %f, %f,' % (target.x, target.y, target.z))
    camera_settings.append('%f, %f, %f,' % (position.x, position.y, position.z))
    camera_settings.append('%f, %f, %f' % (up.x, up.y, up.z))
    camline = '%s' % ' ' .join(camera_settings)
    orcFile.write('%s\n' % camline)

# ---------------------------------------------------------------------------
def writeCameraRange(filepath, octane_render, only_selected=False):

    scene = bpy.context.scene
    frame_start = scene.frame_start
    frame_end = scene.frame_end
    found_it = False

    cameras = []

    for obj in scene.objects:
        if only_selected and not obj.select:
            continue
        if obj.type != 'CAMERA':
            continue
        found_it = True
        cameras.append((obj, obj.data))

    if not found_it:
        return False

    frame_range = range(frame_start, frame_end + 1)
    current_frame = scene.frame_current

    orcFile = open(filepath, 'w')
    for f in frame_range:
        scene.frame_set(f)
        for obj, obj_data in cameras:
            writeOrcLine(obj, orcFile, octane_render)
    orcFile.close()
    scene.frame_current = current_frame
    return found_it

def writeSelectedCamera(filepath, octane_render):
    scene = bpy.context.scene
    cameras = []
    only_selected = True
    found_it = False
    for obj in scene.objects:
        if only_selected and not obj.select:
            continue
        if obj.type != 'CAMERA':
            continue
        found_it = True
        cameras.append((obj, obj.data))

    if not found_it:
        return False

    if octane_render.orc_export_type == 'Current':
        orcFile = open(filepath, 'w')
    else:
        orcFile = open(filepath, 'a')
    for obj, obj_data in cameras:
        writeOrcLine(obj, orcFile, octane_render)
    orcFile.close()
    return found_it

# ---------------------------------------------------------------------------
def getFilePath(self, octane_render):
    export_path = bpath.abspath(octane_render.path)
    if not os.path.isdir(export_path):
        msg = 'The project directory is invalid ("%s")' % (export_path)
        update_status(2,msg)
        notify_user (self)
        return False, ""

    baseName = fixName(octane_render.project_name)
    if baseName == '':
        msg = 'The project name is missing ("%s")' % (baseName)
        update_status(2,msg)
        notify_user (self)
        return False, ""

    filename_ext = ".orc"
    filepath = "".join([export_path, baseName, filename_ext])
    return True, filepath


# ---------------------------------------------------------------------------
### EXPORT RANGE BUTTON ###
class BUTTON_orc_export_range(Operator):
    bl_idname = "ops.orc_export_range"
    bl_label = "Export frame range"
    bl_description = "Exports selected camera for the current frame range to Orc file"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        octane_render = context.scene.octane_render
        success, filepath = getFilePath(self, octane_render)
        if not success:
            return {"CANCELLED"}
        else:
            start = datetime.datetime.now()
            only_selected = BoolProperty(name="Only Selected", default=True)
            
            if not writeCameraRange(filepath, octane_render, only_selected):
                msg = 'No camera selected'
                update_status(2, msg)
                notify_user (self)
                return {"CANCELLED"}

        msg = "Camera export completed in {0})\n".format(datetime.datetime.now() - start)
        self.report({'INFO'}, msg)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

# ---------------------------------------------------------------------------
### EXPORT SELECTED BUTTON ###
class BUTTON_orc_export_selected_camera(Operator):
    bl_idname = "ops.orc_export_selected_camera"
    bl_label = "Export selected camera"
    bl_description = "Export 1 frame of current camera position to Orc file"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        octane_render = context.scene.octane_render
        success, filepath = getFilePath(self, octane_render)
        if not success:
            return {"CANCELLED"}
        else:
            start = datetime.datetime.now()
            only_selected = BoolProperty(name="Only Selected", default=True)
            if not writeSelectedCamera(filepath, octane_render):
                msg = 'No camera selected'
                update_status(2, msg)
                notify_user (self)
                return {"CANCELLED"}

        msg = "Camera export completed"
        self.report({'INFO'}, msg)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

# ---------------------------------------------------------------------------
### EXPORT APPEND BUTTON ###
class BUTTON_orc_append_selected_camera(Operator):
    bl_idname = "ops.orc_append_selected_camera"
    bl_label = "Append selected camera"
    bl_description = "Append 1 frame of current camera position to Orc file"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        octane_render = context.scene.octane_render
        success, filepath = getFilePath(self, octane_render)
        if not success:
            return {"CANCELLED"}
        else:
            start = datetime.datetime.now()
            only_selected = BoolProperty(name="Only Selected", default=True)
            if not writeSelectedCamera(filepath, octane_render):
                msg = 'No camera selected'
                update_status(2, msg)
                notify_user (self)
                return {"CANCELLED"}

        msg = "Camera append completed."
        self.report({'INFO'}, msg)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

# ---------------------------------------------------------------------------
#  This function is provided as a means for executing any of the 3 Orc export
#  functions using a shortcut key. The key is defined in 'User preferences',
#  under 'Input', '3D View', 'Object Mode'. Click 'Add new' and use 
#  'ops.orc_shortcut' as the function, then specify the shortcut key to use.
#  When pressed in 3D view with a selected camera, the currently highlighted
#  Orc Export function will run.
# ---------------------------------------------------------------------------
class BUTTON_orc_shortcut(Operator):
    bl_idname = "ops.orc_shortcut"
    bl_label = "Export Orc file"
    bl_description = "Shortcut to range, current, or append ORC export"

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        octane_render = context.scene.octane_render
        success, filepath = getFilePath(self, octane_render)
        if not success:
            return {"CANCELLED"}
        else:
            start = datetime.datetime.now()
            only_selected = BoolProperty(name="Only Selected", default=True)
            if octane_render.orc_export_type == 'Range':
                start = datetime.datetime.now()
                only_selected = BoolProperty(name="Only Selected", default=True)
                writeCameraRange(filepath, octane_render, only_selected)
                msg = "Camera export completed in {0})\n".format(datetime.datetime.now() - start)
                self.report({'INFO'}, msg)
                return {"FINISHED"}
            else:
                if not writeSelectedCamera(filepath, octane_render):
                    msg = 'No camera selected'
                    update_status(2, msg)
                    notify_user (self)
                    return {"CANCELLED"}
        msg = "Camera export completed."
        self.report({'INFO'}, msg)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


# ---------------------------------------------------------------------------
### ORC PANEL ###
class OBJECT_PT_orc_animation(Panel):
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "render"
    bl_label = "ORC animation export"
    COMPAT_ENGINES = {'OCT_RENDER'}
    
    
    @classmethod
    def poll(self, context):
        rnd = context.scene.render
        return rnd.engine == "OCT_RENDER" and rnd.use_game_engine == False
    
    def draw_header(self, context):
        layout = self.layout
        layout.label(text="", icon="OUTLINER_DATA_CAMERA")
 
    def draw(self, context):
        layout = self.layout
        octane_render = context.scene.octane_render
        row = layout.row()
        row.prop(octane_render, "orc_export_type", expand=True)
        col = layout.column(align=True)
        if octane_render.orc_export_type == 'Range':
            col.operator("ops.orc_export_range")
        elif octane_render.orc_export_type == 'Current':
            col.operator("ops.orc_export_selected_camera")
        else:
            col.operator("ops.orc_append_selected_camera")

def register():
    bpy.utils.register_class(OBJECT_PT_orc_animation)
    bpy.utils.register_class(BUTTON_orc_export_range)
    bpy.utils.register_class(BUTTON_orc_export_selected_camera)
    bpy.utils.register_class(BUTTON_orc_append_selected_camera)
    bpy.utils.register_class(BUTTON_orc_shortcut)

def unregister():
    bpy.utils.unregister_class(BUTTON_orc_shortcut)
    bpy.utils.unregister_class(BUTTON_orc_append_selected_camera)
    bpy.utils.unregister_class(BUTTON_orc_export_selected_camera)
    bpy.utils.unregister_class(BUTTON_orc_export_range)
    bpy.utils.unregister_class(OBJECT_PT_orc_animation)
